﻿// -----------------------------------------------------------------------
// <copyright file="MultiDPIImage.cs" company="Quppa.net">
// Written by David Warner. Use freely.
// </copyright>
// -----------------------------------------------------------------------

namespace Quppa
{
    using System;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.Linq;
    using System.Runtime.InteropServices;
    using System.Windows.Markup;
    using System.Windows.Media;
    using System.Windows.Media.Imaging;

    public class MultiDPIImageSource : MarkupExtension
    {
        private static readonly int DPIX, DPIY, DPI;

        private const int LOGPIXELSX = 88, LOGPIXELSY = 90;

        [DllImport("user32.dll")]
        private static extern IntPtr GetDC(IntPtr hWnd);

        [DllImport("gdi32.dll")]
        private static extern int GetDeviceCaps(IntPtr hdc, int nIndex);

        private readonly Uri imageuri;

        public MultiDPIImageSource(string path)
        {
            imageuri = new Uri("pack://application:,,,/" + path);
        }

        static MultiDPIImageSource()
        {
            DPIX = GetDeviceCaps(GetDC(IntPtr.Zero), LOGPIXELSX);
            DPIY = GetDeviceCaps(GetDC(IntPtr.Zero), LOGPIXELSY);
            DPI = GetDPI(DPIX, DPIY);
        }

        private static int GetDPI(double dpix, double dpiy)
        {
            return GetDPI((int)Math.Round(dpix), (int)Math.Round(dpiy));
        }

        private static int GetDPI(int dpix, int dpiy)
        {
            return Math.Max(dpix, dpiy);
        }

        private BitmapFrame GetBestFrame(ReadOnlyCollection<BitmapFrame> frames)
        {
            SortedList<int, BitmapFrame> sortedframes = new SortedList<int, BitmapFrame>();
            foreach (BitmapFrame frame in frames)
                if (!sortedframes.ContainsKey(GetDPI(frame.DpiX, frame.DpiY)))
                    sortedframes.Add(GetDPI(frame.DpiX, frame.DpiY), frame);

            foreach (var frame in sortedframes)
            {
                if (frame.Key == DPI)
                    return frame.Value;

                if (DPI < frame.Key)
                    return frame.Value;
            }

            return sortedframes.Last().Value;
        }

        public override object ProvideValue(IServiceProvider serviceProvider)
        {
            try
            {
                TiffBitmapDecoder decoder = new TiffBitmapDecoder(
                    this.imageuri, BitmapCreateOptions.None, BitmapCacheOption.Default);
                BitmapFrame bestframe = this.GetBestFrame(decoder.Frames);

                return bestframe;
            }
            catch
            {
                return null;
            }
        }

        public BitmapFrame GetBestFrame()
        {
            try
            {
                TiffBitmapDecoder decoder = new TiffBitmapDecoder(this.imageuri, BitmapCreateOptions.None, BitmapCacheOption.Default);
                BitmapFrame bestframe = this.GetBestFrame(decoder.Frames);

                return bestframe;
            }
            catch
            {
                return null;
            }
        }
    }

    public class MultiDPIImageScalingMode : MarkupExtension
    {
        private static readonly int DPIX, DPIY, DPI;

        private const int LOGPIXELSX = 88, LOGPIXELSY = 90;

        [DllImport("user32.dll")]
        private static extern IntPtr GetDC(IntPtr hWnd);

        [DllImport("gdi32.dll")]
        private static extern int GetDeviceCaps(IntPtr hdc, int nIndex);

        private readonly Uri imageuri;

        private readonly BitmapScalingMode fallbackscalingmode;

        public MultiDPIImageScalingMode(string path)
        {
            imageuri = new Uri("pack://application:,,,/" + path);
            this.fallbackscalingmode = BitmapScalingMode.Linear;
        }

        public MultiDPIImageScalingMode(string path, BitmapScalingMode scalingmode)
        {
            imageuri = new Uri("pack://application:,,,/" + path);
            this.fallbackscalingmode = scalingmode;
        }

        static MultiDPIImageScalingMode()
        {
            DPIX = GetDeviceCaps(GetDC(IntPtr.Zero), LOGPIXELSX);
            DPIY = GetDeviceCaps(GetDC(IntPtr.Zero), LOGPIXELSY);
            DPI = GetDPI(DPIX, DPIY);
        }

        private static int GetDPI(double dpix, double dpiy)
        {
            return GetDPI((int)Math.Round(dpix), (int)Math.Round(dpiy));
        }

        private static int GetDPI(int dpix, int dpiy)
        {
            return Math.Max(dpix, dpiy);
        }

        private BitmapScalingMode GetScalingMode(ReadOnlyCollection<BitmapFrame> frames)
        {
            SortedList<int, BitmapFrame> sortedframes = new SortedList<int, BitmapFrame>();
            foreach (BitmapFrame frame in frames)
                if (!sortedframes.ContainsKey(GetDPI(frame.DpiX, frame.DpiY)))
                    sortedframes.Add(GetDPI(frame.DpiX, frame.DpiY), frame);

            foreach (var frame in sortedframes)
            {
                if (frame.Key == DPI) return BitmapScalingMode.NearestNeighbor;

                if (DPI < frame.Key) return fallbackscalingmode;
            }

            return fallbackscalingmode;
        }

        public override object ProvideValue(IServiceProvider serviceProvider)
        {
            try
            {
                TiffBitmapDecoder decoder = new TiffBitmapDecoder(
                    this.imageuri, BitmapCreateOptions.None, BitmapCacheOption.Default);
                BitmapScalingMode scalingmode = this.GetScalingMode(decoder.Frames);

                return scalingmode;
            }
            catch
            {
                return BitmapScalingMode.Unspecified;
            }
        }

        public BitmapScalingMode GetScalingMode()
        {
            try
            {
                TiffBitmapDecoder decoder = new TiffBitmapDecoder(
                    this.imageuri, BitmapCreateOptions.None, BitmapCacheOption.Default);
                BitmapScalingMode scalingmode = this.GetScalingMode(decoder.Frames);

                return scalingmode;
            }
            catch
            {
                return BitmapScalingMode.Unspecified;
            }
        }
    }
}
